import sys

# Probably not the final strategy, but start with some hardcoded spacing values
font_info = {
    'Arial': {'regular': [16, 17, 18, 19, 22, 23, 24, 26, 27], 'bold': [16, 18, 19, 19, 22, 24, 24, 27, 29]},
    'Tahoma': [16, 18, 19, 21, 23, 24],
    'Kozuka Gothic Pro H': [19, 22, 23, 24, 27, 29],
    'Helvetica LT std': [15, 18, 19, 20, 23, 24],
    'Open Sans Semibold': [19, 22, 23, 24, 27, 28],
    'Franklin Gothic Book': [17, 20, 21, 21, 24, 25],
    'Franklin Gothic Demi': [17, 20, 21, 21, 24, 25],
    'Franklin Gothic Medium': [17, 20, 21, 21, 24, 25],
    'MS Gothic': [13, 15, 16, 17, 19, 20, 21, 23, 24],
    'Gadugi': {'regular': [16, 18, 19, 20, 22, 24, 25], 'bold': [16, 18, 19, 20, 21, 24, 25]},
    'Verdana': [16, 18, 18, 20, 23, 25, 25]
}

# TODO: object with spacing property
#def spacing(font, size, bold=False):
def spacing(style):
    global __warned__
    font = style.fontname
    size = style.fontsize
    bold = 'B' in style.fontstyle
    spacing = None
    if font in font_info:
        if type(cur := font_info[font]) is dict:
            cur = cur['bold' if bold else 'regular']
        if 0 <= size - 10 < len(cur):
            spacing = cur[size - 10]
    if spacing is None:
        raise KeyError(f'Font "{font}", size {size}{" (bold)" if bold else ""} not known. Disable the experimental spacing feature to process the file.')
    return spacing



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__version__ = '0.0.15'

__all__ = ['KBPFile', 'AssConverter', 'DoblonTxtConverter', 'LRCConverter']

from .kbp import *
from .doblontxt import *
from .lrc import *
from .converters import *



import re
import charset_normalizer

class DoblonTxt:
    def __init__(self, txtfile: str):
        lines = []
        line = []
        # TODO: look into only opening the file once
        with open(txtfile, 'r', encoding=charset_normalizer.from_path(txtfile).best().encoding) as f:
            for syl in f:
                start, stop, text = syl.rstrip("\n").split('-', maxsplit=2)
                while text.startswith("\\n"):
                    lines.append(line)
                    line = []
                    text = text[2:]
                line.append((self.time_to_ms(start),self.time_to_ms(stop),text))
                while text.endswith("\\n"):
                    text = text[:-2]
                    line[-1]=(self.time_to_ms(start), self.time_to_ms(stop), text)
                    lines.append(line)
                    line = []
        self.lines = lines

    @staticmethod
    def time_to_ms(timestamp: str) -> int:
        r = re.match(r'(\d+:)?(\d+)\.(\d+)', timestamp)
        return int(r.group(3)) + 1000 * int(r.group(2)) + (60 * 1000 * int(r.group(1)[:-1]) if r.group(1) else 0)



import collections
import re
import string
import typing
import io
import warnings
import os.path
from . import validators

class KBPFile:

    DIVIDER = "-----------------------------"
    HEADER_START = "\n".join((
        DIVIDER,
        "KARAOKE BUILDER STUDIO",
        "www.KaraokeBuilder.com",
        "",
        DIVIDER,
        "HEADERV2",
        "",
        "'--- Template Information ---",
        "\n"
    ))

    def __init__(self, kbpFile=None, **kwargs):
        self.pages = []
        self.images = []
        if kbpFile is None:
            self.colors = KBPPalette("055","FFF","000","E70","940","CFF","033","0DD","077","FCF","303","F3F","818","000","FFF","000")
            self.styles = KBPStyleCollection({1: KBPStyle(style_no=1, name='Default', textcolor=1, outlinecolor=2, textwipecolor=3, outlinewipecolor=4, fontname='Arial', fontsize=12, fontstyle='B', charset=0, outlines=[2, 2, 2, 2], shadows=[0, 0], wipestyle=0, allcaps='L', fixed=False), 2: KBPStyle(style_no=2, name='Male', textcolor=5, outlinecolor=6, textwipecolor=7, outlinewipecolor=8, fontname='Arial', fontsize=12, fontstyle='B', charset=0, outlines=[2, 2, 2, 2], shadows=[0, 0], wipestyle=0, allcaps='L', fixed=False), 3: KBPStyle(style_no=3, name='Female', textcolor=9, outlinecolor=10, textwipecolor=11, outlinewipecolor=12, fontname='Arial', fontsize=12, fontstyle='B', charset=0, outlines=[2, 2, 2, 2], shadows=[0, 0], wipestyle=0, allcaps='L', fixed=False), 4: KBPStyle(style_no=4, name='Other', textcolor=4, outlinecolor=8, textwipecolor=12, outlinewipecolor=14, fontname='Arial', fontsize=12, fontstyle='B', charset=0, outlines=[2, 2, 2, 2], shadows=[0, 0], wipestyle=0, allcaps='L', fixed=False)})
            self.margins = {'left': 2, 'right': 2, 'top': 7, 'spacing': 12}
            self.other = {'bordercolor': 0, 'wipedetail': 2}
            return
        needsclosed = False
        if not isinstance(kbpFile, io.IOBase):
            self.filename = kbpFile
            kbpFile = open(kbpFile, "r", encoding="utf-8")
            needsclosed = True
        self.parse([x.rstrip() for x in kbpFile.readlines()], **kwargs)
        if needsclosed:
            kbpFile.close()

    def parse(self, kbpLines, resolve_colors=False, resolve_wipe=True, template=None):
        in_header = False
        divider = False
        for x, line in enumerate(kbpLines):
            if in_header:
                if line.startswith("'Palette Colours"):
                    self.colors = KBPPalette.from_string(kbpLines[x+1])
                elif line.startswith("'Styles"):
                    data = kbpLines[x+1:kbpLines.index("  StyleEnd", x+1)]
                    opts = {"palette": self.colors} if resolve_colors else {}
                    self.styles = KBPStyleCollection.from_textlines([x for x in data if not x.startswith("'")], **opts)
                elif line.startswith("'Margins"):
                    self.parse_margins(kbpLines[x+1])
                elif line.startswith("'Other"):
                    self.parse_other(kbpLines[x+1])
                elif line == "'--- Track Information ---":
                    if template:
                        return
                    data = kbpLines[x+1:kbpLines.index(KBPFile.DIVIDER, x+1)]
                    self.parse_trackinfo(data)
                    if self.trackinfo["Status"] != '1':
                        raise NotImplementedError("Tracks must be synced before they can be used with kbputils.")

            elif divider and line == "PAGEV2":
                data = kbpLines[x+1:kbpLines.index(KBPFile.DIVIDER, x+1)]
                opts = {"default_wipe": self.other['wipedetail']} if resolve_wipe else {}
                self.pages.append(KBPPage.from_textlines(data, **opts))

            elif divider and line == "IMAGE":
                # TODO: Determine if it's ever possible to have multiple image lines in one section
                data = kbpLines[x+1]
                self.images.append(KBPImage.from_string(data))

            if divider and line == "HEADERV2":
                in_header = True

            if line == KBPFile.DIVIDER:
                in_header = False
                divider = True
            # Ignore empty/comment lines and still consider the previous line to be a divider
            elif line != "" and not line.startswith("'"):
                divider = False

        missing = ', '.join(filter(lambda x: not hasattr(self, x), ('colors', 'styles', 'margins', 'other','pages')))
        if missing:
            raise ValueError(f"Invalid KBP file, missing sections: {missing}")

        if not hasattr(self, 'trackinfo') and template == False: # ignore when None
            raise ValueError("Invalid KBP file, missing track info. If this was intended to be used as a template, set template to True or None")

    

    def parse_margins(self, margin_line):
        self.margins = dict(zip(("left", "right", "top", "spacing"), (int(x) for x in margin_line.strip().split(","))))

    def parse_other(self, other_line):
        self.other = dict(zip(("bordercolor", "wipedetail"), (int(x) for x in other_line.strip().split(","))))

    # This is formatted in a way that it may allow freeform key/value pairs, so
    # leaving this as just a dict, but so far it always seems to contain exactly:
    # - Status
    # - Title
    # - Artist
    # - Audio
    # - BuildFile
    # - Intro
    # - Outro
    # - Comments
    def parse_trackinfo(self, trackinfo_lines):
        self.trackinfo = {}
        prev = None
        for line in trackinfo_lines:
            if line.startswith(" "):
                self.trackinfo[prev] += f"\n{line.lstrip()}"
            elif line != "" and not line.startswith("'"):
                fields = line.split(maxsplit=1)
                if len(fields) == 1:
                    fields.append("")
                self.trackinfo[fields[0]] = fields[1]
                prev = fields[0]

    

    # Convenience method to get all the lyric text without timing info,
    # potentially with syllable marks added. To get lyric text that could be used
    # to restart sync on a new project, use:
    # myKBPFile.text(include_empty=True, syllable_separator="/", space_is_separator=True)
    def text(self, page_separator="", include_empty=False, syllable_separator="", space_is_separator=False):
        result = []
        for page in self.pages:
            lines = []
            for line in page.lines:
                if include_empty or not line.isempty():
                    lines.append(line.text(syllable_separator=syllable_separator, space_is_separator=space_is_separator))
            result.append("\n".join(lines))
        return f"\n{page_separator}\n".join(result)

    def writeFile(self, kbpFile, allow_overwrite=False):
        if any(x.has_colors() for x in self.styles.values()):
            raise ValueError("Unable to write styles when they contain arbitrary colors. They must use palette indexes.")
        if not isinstance(kbpFile, io.IOBase):
            if not allow_overwrite and os.path.exists(kbpFile) and hasattr(self,'filename') and os.path.samefile(kbpFile, self.filename):
                raise ValueError("Refusing to write back to original filename. Set allow_overwrite if you need to do so.")
            kbpFile = open(kbpFile, "w", encoding="utf-8",newline="\r\n")
            needsclosed = True
        else:
            needsclosed = False
        kbpFile.write(KBPFile.HEADER_START)
        kbpFile.write(self.colors.toKBP())
        kbpFile.write(self.styles.toKBP())
        kbpFile.write("\n  ".join((
            "'Margins : L,R,T,Line Spacing",
            ",".join(str(x) for x in self.margins.values())
        )) + "\n\n")
        kbpFile.write("\n  ".join((
            "'Other: Border Colour,Detail Level",
            ",".join(str(x) for x in self.other.values())
        )) + "\n\n")

        if hasattr(self, 'trackinfo'):
            kbpFile.write("'--- Track Information ---\n\n")
            for x in self.trackinfo:
                if x == 'Comments':
                    kbpFile.write("\n")
                tmp = ('\n' + ' ' * 10).join(self.trackinfo[x].split('\n'))
                kbpFile.write(f"{x:<10}{tmp}\n")
            kbpFile.write("\n")
            #kbpFile.write("\n".join(f"""{'\n' if x == 'Comments' else ''}{x:<10}{
            #    ('\n' + ' ' * 10).join(self.trackinfo[x].split('\n'))
            #}""" for x in self.trackinfo) + "\n\n")
            for page in self.pages:
                kbpFile.write(page.toKBP())
            for image in self.images:
                kbpFile.write(image.toKBP())
        # else template

        kbpFile.write(KBPFile.DIVIDER + "\n\n")

        if needsclosed:
            kbpFile.close()

# I didn't want to use a list because it can't have specific size, and liked
# all the data being tuple-compatible, but still wanted list-like access with [].
# There are modules that make something like a "frozen" list, but I didn't want
# to pull in any more extra modules, so you get this monstrosity, sorry world
@validators.validated_instantiation
class KBPPalette(collections.namedtuple("KBPPalette", tuple(range(16)), rename=True)):

    __annotations__ = dict((f"_{x}",str) for x in range(16))
    
    # namedtuple-generated classes don't have __init__ so must instead override __new__
    def __new__(cls, *colors):
        assert len(colors) == 16 and all(re.match(r"[0-9A-F]{3}$", x) for x in colors)
        return super().__new__(cls, *colors)

    # Hide the field names because they are just _0 to _15 and aren't going to be used directly
    def __repr__(self):
        return "KBPPalette(" + ", ".join(repr(x) for x in self) + ")"

    def __getitem__(self, x):
        if isinstance(x, int) and 0 <= x < 16:
            return getattr(self, f"_{x}")
        else:
            raise KeyError(x)

    # Set available colors to what is configured in the palette
    @staticmethod
    def from_string(palette_line):
        return KBPPalette(*palette_line.lstrip().split(","))

    def toKBP(self):
        return f"'Palette Colours (0-15)\n  {','.join(self)}\n\n"

    def as_rgb24(self):
        return ["".join(y * 2 for y in x) for x in self]

    def as_rgba32(self):
        return [x+"FF" for x in self.as_rgb24()]

@validators.validated_instantiation
class KBPLineHeader(typing.NamedTuple):
    align: str
    style: str
    start: int
    end: int
    right: int
    down: int
    rotation: int

    def isfixed(self):
        return self.style.islower()

    def toKBP(self):
        return "/".join(str(x) for x in self)

@validators.validated_instantiation
class KBPSyllable(typing.NamedTuple):
    syllable: str
    start: int
    end: int
    wipe: int

    def isempty(self):
        return self.syllable == ""

    def isprogressive(self):
        # Zero means unresolved default wipe, so result is undefined. Otherwise
        # everything less than 5 is progressive with varying level of wipe detail
        return None if self.wipe == 0 else (self.wipe < 5)

    def toKBP(self):
        return f"{self.syllable + '/':<15}" + "/".join(str(x) for x in self[1:])

class KBPLine(typing.NamedTuple):
    header: KBPLineHeader
    syllables: list

    # There's only one header, so may as well pass anything unresolved down to it
    def __getattr__(self, attr):
        return getattr(self.header, attr)

    # These will just return "can't set" errors since it's immutable, but that
    # makes more sense than "has no attribute"
    def __setattr__(self, attr, val):
        return setattr(self.header, attr, val)

    def text(self, syllable_separator="", space_is_separator=False):

        # Special case for empty lines because when importing lyrics into KBS,
        # a syllable separator by itself is used to denote an empty line (as
        # opposed to a page break)
        if self.isempty():
            return syllable_separator

        # When space is considered a separator, underscore is used to represent
        # a non-splitting space
        elif space_is_separator and syllable_separator != "":
            result = ""
            for syl in self.syllables:
                syltext = syl.syllable
                syltext = re.sub(r"( +)(?=[^ ])", lambda m: "_" * len(m.group(1)), syltext)
                result += syltext
                if not syltext.endswith(" "):
                    result += syllable_separator
            return result[:-len(syllable_separator)]

        # Basic case - output syllables split only by the separator character(s)
        else:
            return syllable_separator.join(x.syllable for x in self.syllables)

    def isempty(self):
        return not self.syllables or (len(self.syllables) == 1 and self.syllables[0].isempty())

    def toKBP(self):
        return "\n".join((self.header.toKBP(), *(x.toKBP() for x in self.syllables))) + "\n"

@validators.validated_instantiation
class KBPStyle(typing.NamedTuple):
    style_no: int
    name: str
    textcolor: str | int
    outlinecolor: str | int
    textwipecolor: str | int
    outlinewipecolor: str | int
    fontname: str
    fontsize: int
    fontstyle: str
    charset: int
    outlines: typing.Iterable
    shadows: typing.Iterable
    wipestyle: int
    allcaps: str
    fixed: bool

    def has_colors(self):
        fields = ("textcolor", "outlinecolor", "textwipecolor", "outlinewipecolor")
        if all(isinstance(getattr(self,x), str) for x in fields):
            return True
        elif all(isinstance(getattr(self,x), int) for x in fields):
            return False
        else:
            raise TypeError("Mixed/unexpected types found in color parameters:\n\t" + 
                "\n\t".join([": ".join((x, str(getattr(self,x)))) for x in fields]))

    def resolve_colors(style, palette):
        if style.has_colors():
            warnings.warn(f"Colors were already resolved on style {name}. Palette may not be used as intended", RuntimeWarning)
            return style
        else:
            fields = ("textcolor", "outlinecolor", "textwipecolor", "outlinewipecolor")
            return style._replace(**dict((x, palette[getattr(style,x)]) for x in fields))

    # Create a "fixed" version of a style (no wiping)
    # Technically the wipe colors are still defined with their original values,
    # but they are not used, so it makes sense to redefine them to their
    # non-wiped values for compatibility with other formats
    def make_fixed(style):
        if style.fixed:
            return style
        else:
            return style._replace(
                name = style.name + "_fixed",
                style_no = -style.style_no,
                textwipecolor = style.textcolor,
                outlinewipecolor = style.outlinecolor,
                fixed = True)

    def toKBP(self):
        tmp = tuple(str(x) for x in self)
        return "\n    ".join((
            ",".join((f"  Style{(self.style_no - 1):02d}",) + tmp[1:6]),
            ",".join(tmp[6:10]),
            ",".join(tuple(str(x) for x in self.outlines) + tuple(str(x) for x in self.shadows) + tmp[12:14])
        )) + "\n\n"

class KBPStyleCollection(dict):
    __slots__ = ()

    HEADER = "\n".join((
        "'Styles (00-19)",
        "'  Number,Name",
        "'  Colour: Text,Outline,Text Wipe,Outline Wipe",
        "'  Font  : Name,Size,Style,Charset",
        "'  Other : Outline*4,Shadow*2,Wiping,Uppercase",
        "\n"
    ))

    def __repr__(self):
        return "KBPStyleCollection(" + super().__repr__() + ")"

    # Alias style letters used in line headers to style numbers used in definition
    @staticmethod
    def alpha2key(alpha):
        if alpha in list(string.ascii_uppercase):
            return string.ascii_uppercase.index(alpha)+1
        elif alpha in list(string.ascii_lowercase):
            return -string.ascii_lowercase.index(alpha)-1

    def __missing__(self, key):
        if key in list(string.ascii_letters):
            return self[KBPStyleCollection.alpha2key(key)]

        # Auto-vivify fixed styles
        elif isinstance(key, int) and -key in self.keys():
            self[key] = self[-key].make_fixed()
            return self[key]

        # Undefined styles should return the default style
        # The caller can determine this happened if needed by
        # comparing style.style_no with the requested index
        # A converter can choose on its side whether it wants
        # the undefined style to be considered a new one or 
        # reuse the name of the default
        elif isinstance(key, int) and (1 <= abs(key) <= 26):
            return self[1] if key > 0 else self[-1]

        else:
            raise KeyError(key)

    # Hopefully the checks in the next several methods will be enough to keep bad data out...
    @validators.validated_types
    @staticmethod
    def __assert_valid_item(key: int, value: KBPStyle):
        if not (1 <= key <= 26 or -26 <= key <= -1):
            raise KeyError(f"Invalid index for item in KBPStyleCollection: {key}")

    # Using *args is the only way to get unnamed positional arguments.
    # Otherwise if the argument name happens to be a key in the kwargs, it causes
    # a duplicate argument. The decorator validates that it still only gets one
    # argument. As far as I know this is the only way to properly replicate the
    # signatures of the dict methods that can take an object *or* a mapping via
    # kwargs
    @validators.validated_structures(__assert_valid_item)
    @validators.one_arg
    def __init__(self, *args, **kwargs):
        arg = args[0] if args else []
        super().__init__(arg, **kwargs)
    
    def __setitem__(self, key, value):
        KBPStyleCollection.__assert_valid_item(key, value)
        super().__setitem__(key, value)
    
    @validators.validated_structures(__assert_valid_item)
    @validators.one_arg
    def update(self, *args, **kwargs):
        arg = args[0] if args else []
        super().update(arg, **kwargs)

    @validators.validated_structures(__assert_valid_item)
    def __ior__(self, arg):
        return super().__ior__(arg)

    def __or__(self, arg):
        # Data validated by __init__
        return KBPStyleCollection(super().__or__(arg))

    @staticmethod
    def fromkeys(keys, value=None):
        # Data validated by __init__
         return KBPStyleCollection(dict.fromkeys(keys, value))

    # Set available styles based on the configuration
    @staticmethod
    def from_textlines(style_lines, palette=None):
        styles = KBPStyleCollection()
        fields = {}
        style_no = None
        for n, line in enumerate(style_lines):
            line = line.lstrip()
            if line == "" and style_no is not None:
                style_no = None
                fields = {}
            elif style_no is None and line.startswith("Style"):
                tmp = line.split(",")
                style_no = int(tmp[0][5:]) + 1
                fields['style_no'] = style_no
                tmp = [tmp[1], *(int(x) for x in tmp[2:])]
                fields.update(dict(zip(("name", "textcolor", "outlinecolor", "textwipecolor", "outlinewipecolor"),tmp)))
                tmp = style_lines[n+1].lstrip().split(",")
                tmp[1] = int(tmp[1])
                tmp[3] = int(tmp[3])
                fields.update(dict(zip(("fontname", "fontsize", "fontstyle", "charset"),tmp)))
                tmp = style_lines[n+2].lstrip().split(",")
                tmp[:-1] = [int(x) for x in tmp[:-1]]
                fields.update(dict(zip(("outlines", "shadows", "wipestyle", "allcaps", "fixed"),(tmp[:4],tmp[4:6],tmp[6],tmp[7],False))))
                result = KBPStyle(**fields)
                if palette:
                    result = result.resolve_colors(palette)
                # Adding 1 to style for 2 reasons:
                # - Style 0/A is shown in the KBS UI as 01
                # - It allows indexing fixed styles as negative numbers
                styles[style_no] = result
            # else second/third line of styles already processed
        return styles
    
    
    def key2alpha(key: int):
        if key < 0:
            return string.ascii_lowercase[-key-1]
        else:
            return string.ascii_uppercase[key-1]

    # This is almost definitely ill-advised, but this is like dict.keys() and
    # returns a view object for the alpha aliases of the keys. And of course it
    # can give you an iterator so you can do something like:
    # for x in kbpfile.styles.alpha_keys():
    #     ...
    def alpha_keys(self):
        class alphakey_view:
            def __init__(slf, obj):
                slf.__obj = obj

            def __len__(slf):
                return len(slf.__obj)

            def __reversed__(slf):
                return alphakey_view(reversed(slf.__obj))

            def __repr__(slf):
                return f"style_alphakey_view({', '.join(repr(KBPStyleCollection.key2alpha(x)) for x in slf.__obj)})"

            def __iter__(slf):
                class alphakey_iterator:
                    def __init__(sf, it):
                        sf.__iter = it
                    def __repr__(sf):
                        return f"<style_alphakey_iterator at {hex(id(sf))}>"
                    def __next__(sf):
                        return next(sf.__iter)
                return alphakey_iterator(KBPStyleCollection.key2alpha(x) for x in slf.__obj)

        return alphakey_view(self.keys())

    def alpha_iter(self):
        return iter(self.alpha_keys())

    def toKBP(self):
        return KBPStyleCollection.HEADER + "".join(x.toKBP() for x in self.values() if x.style_no > 0) + "  StyleEnd\n\n"
        

@validators.validated_instantiation
class KBPPage(typing.NamedTuple):
    remove: str
    display: str
    lines: list

    @staticmethod
    def from_textlines(page_lines, default_wipe = None):
        lines=[]
        syllables=[]
        header=None
        transitions=["", ""] # Default line by line
        for x in page_lines:
            if header is None and re.match(r"[LCR]/[a-zA-Z](/\d+){2}(/-?\d+){3}$", x): # Only last 3 fields can be negative
                fields = x.split("/")
                fields[2:] = [int(y) for y in fields[2:]]
                header = KBPLineHeader(**dict(zip(("align", "style", "start", "end", "right", "down", "rotation"), fields)))
            elif x == "" and header is not None:
                # Handle previous line
                lines.append(KBPLine(header=header, syllables=syllables))
                syllables = []
                header = None
            elif header is None and x.startswith("FX/"):
                transitions = x.split('/')[1:]
            elif x != "":
                fields = x.split("/")
                fields[0] = re.sub(r"{-}", "/", fields[0]) # This field uses this as a surrogate for / since that denotes end of syllable
                fields[1] = fields[1].lstrip() # Only the second field should have extra spaces
                fields[1:] = [int(y) for y in fields[1:]]
                if default_wipe and fields[3] == 0:
                    fields[3] = default_wipe
                syllables.append(KBPSyllable(**dict(zip(("syllable", "start", "end", "wipe"), fields))))
        return KBPPage(*transitions, lines)

    def get_start(self):
        return min(line.start for line in self.lines if not line.isempty())

    def get_end(self):
        return max(line.end for line in self.lines)

    def toKBP(self):
        transition = (f"FX/{self.remove}/{self.display}\n",) if (self.remove or self.display) else ()
        return "\n".join((KBPFile.DIVIDER, "PAGEV2", *transition, *(x.toKBP() for x in self.lines))) + "\n"

# Perhaps this could be better named as slideshow, but calling it IMAGE as that's what the header is called in the .kbp
@validators.validated_instantiation
class KBPImage(typing.NamedTuple):
    start: int
    end: int
    filename: str
    leaveonscreen: int

    @staticmethod
    def from_string(image_line):
        fields = image_line.split("/")
        for x in (0, 1, 3):
            fields[x] = int(fields[x])
        return KBPImage(**dict(zip(("start", "end", "filename", "leaveonscreen"),fields)))

    def toKBP(self):
        return "\n".join((KBPFile.DIVIDER, "IMAGE", "/".join(str(x) for x in self))) + "\n\n"



import os
import sys


def _fetch_offset():
    if os.environ.get('APPDATA') and os.path.exists(p := os.path.join(os.environ['APPDATA'], 'Karaoke Builder', 'data_studio.ini')):
        try:
            with open(p, 'r') as f:
                for line in f:
                    field, val = line.split(maxsplit=1)
                    if field == "setoffset":
                        return int(val)
        except OSError:
            pass
    return 0

# Lazy-load the offset attribute so the file is only read if needed

def __getattr__(attr):
    if attr == "offset":
        setattr(sys.modules[__name__], attr, _fetch_offset())
        return getattr(sys.modules[__name__], attr)
    else:
        raise AttributeError(f"Module {__name__!r} has no attribute {attr!r}")



import re
import charset_normalizer

class LRC:
    def __init__(self, lrcfile: str):
        self.lines = []
        self.tags = {}
        # TODO: look into only opening the file once
        with open(lrcfile, 'r', encoding=charset_normalizer.from_path(lrcfile).best().encoding) as f:
            for lrcline in f:
                lrcline = lrcline.rstrip("\r\n")
                if re.fullmatch(r'\[\d{2}:\d{2}.\d{2}\]\s+(<\d{2}:\d{2}.\d{2}>[^<>]*)+<\d{2}:\d{2}.\d{2}>', lrcline):
                    # Ignore the line start times for now - they aren't usually going to be helpful when redoing
                    # layout anyway and some programs don't set them to good values (e.g. KBS LRC export)
                    syls = re.findall(r'<(\d{2}):(\d{2}).(\d{2})>([^<>]*)', lrcline)
                    self.lines.append([(self.time_to_ms(*syls[i][:3]), self.time_to_ms(*syls[i+1][:3]), syls[i][3]) for i in range(len(syls)-1)])
                # For some reason karlyriceditor does [..:..:..]WORD <..:..:..>WORD <..:..:..>
                elif re.fullmatch(r'\[\d{2}:\d{2}.\d{2}\]([^<>]*<\d{2}:\d{2}.\d{2}>)+', lrcline):
                    syls = re.findall(r'[<\[](\d{2}):(\d{2}).(\d{2})[>\]]([^<>]*)', lrcline)
                    self.lines.append([(self.time_to_ms(*syls[i][:3]), self.time_to_ms(*syls[i+1][:3]), syls[i][3]) for i in range(len(syls)-1)])
                elif res := re.fullmatch(r'\[([^\[\]]+)\s*:([^\[\]]+)\]', lrcline):
                    self.tags[res.group(1)] = res.group(2)
                # I don't think this is standard, but it seems to be used as a page break some places
                elif lrcline == '':
                    if self.lines[-1] != []:
                        self.lines.append([])
                else:
                    raise ValueError(f"Invalid LRC line encountered:\n{lrcline}")
            if 'offset' in self.tags:
                offset = int(self.tags.pop('offset'))
                for line in self.lines:
                    for i in range(len(line)):
                        line[i] = (line[i][0] - offset, line[i][1] - offset, line[i][2])

    @staticmethod
    def time_to_ms(m: str, s: str, cs: str) -> int:
        return int(cs)*10 + 1000*int(s) + 60*1000*int(m)



from . import kbp
from . import doblontxt
from . import lrc
from . import converters
from . import __version__
import argparse
import dataclasses
import io
import sys
import collections

def convert_file():
    parser = argparse.ArgumentParser(prog='KBPUtils', description="Various utilities for .kbp files", argument_default=argparse.SUPPRESS)

    parser_data = {
        'kbp2ass': {
            'add_parser': {
                'description': 'Convert .kbp to .ass file',
                'argument_default': argparse.SUPPRESS
            },
            'input': kbp.KBPFile,
            'output': lambda source, args, dest: converters.AssConverter(source, **vars(args)).ass_document().dump_file(dest),
            'output_opts': {
                'encoding': 'utf_8_sig'
            },
            'options': converters.AssOptions
        },
        'doblontxt2kbp': {
            'add_parser': {
                'description': 'Convert Doblon full timing .txt file to .kbp',
                'argument_default': argparse.SUPPRESS
            },
            'input': doblontxt.DoblonTxt,
            'output': lambda source, args, dest: converters.DoblonTxtConverter(source, **vars(args)).kbpFile().writeFile(dest),
            'output_opts': {
                'encoding': 'utf-8',
                'newline': '\r\n'
            },
            'options': converters.DoblonTxtOptions
        },
        'lrc2kbp': {
            'add_parser': {
                'description': 'Convert Enhanced .lrc to .kbp',
                'argument_default': argparse.SUPPRESS
            },
            'input': lrc.LRC,
            'output': lambda source, args, dest: converters.LRCConverter(source, **vars(args)).kbpFile().writeFile(dest),
            'output_opts': {
                'encoding': 'utf-8',
                'newline': '\r\n'
            },
            'options': converters.LRCOptions
        }
    }

    subparsers = parser.add_subparsers(dest='subparser', required=True)

    for p in parser_data:
        cur = subparsers.add_parser(p, **parser_data[p]['add_parser'])

        for field in dataclasses.fields(parser_data[p]['options']):
            name = field.name.replace("_", "-")

            additional_params = {}
            if field.type == int | bool:
                additional_params["type"] = int_or_bool 
            elif hasattr(field.type, "__members__") and hasattr(field.type, "__getitem__"):
                # Handle enum types
                additional_params["type"] = field.type.__getitem__
                additional_params["choices"] = field.type.__members__.values()
            else:
                additional_params["type"] = field.type

            cur.add_argument(
                f"--{name}",
                gen_shortopt(p, name),
                dest = field.name,
                help = (field.type.__name__ if hasattr(field.type, '__name__') else repr(field.type)) + f" (default: {field.default})",
                action = argparse.BooleanOptionalAction if field.type == bool else 'store',
                **additional_params,
            )

        cur.add_argument("--version", "-V", action="version", version=__version__)
        cur.add_argument("source_file")
        cur.add_argument("dest_file", nargs='?')

    args = parser.parse_args()
    subparser = args.subparser
    del args.subparser
    source = parser_data[subparser]['input'](sys.stdin if args.source_file == "-" else args.source_file)
    del args.source_file
    dest = open(args.dest_file, 'w', **parser_data[subparser]['output_opts']) if hasattr(args, 'dest_file') else sys.stdout
    if hasattr(args, 'dest_file'):
        del args.dest_file
    parser_data[subparser]['output'](source, args, dest)

# Auto-generate short option based on field name
used_shortopts=collections.defaultdict(lambda: set("hV"))
def gen_shortopt(command, longopt):
    # Options with - likely have duplication, so use a letter from after the
    # last one
    if len(parts := longopt.split("-")) > 1:
        return gen_shortopt(command, parts[-1])
    for char in longopt:
        if char not in used_shortopts[command]:
            used_shortopts[command].add(char)
            return f"-{char}"

# Coerce a string value into a bool or int
# Accept true|false (case-insensitive), otherwise try int
def int_or_bool(strVal):
    if strVal.upper() == 'FALSE':
        return False
    elif strVal.upper() == 'TRUE':
        return True
    else:
        return int(strVal)

if __name__ == "__main__":
    convert_file()



import ass
import dataclasses
import datetime
import enum
import types
import typing
import collections
import re
import math
from . import kbp
from . import validators
from . import kbs
from . import kbpfont
from . import doblontxt
from . import lrc

class AssAlignment(enum.Enum):
    DEFAULT = 0
    BOTTOM_LEFT = 1
    BOTTOM_CENTER = 2
    BOTTOM_RIGHT = 3
    MIDDLE_LEFT = 4
    MIDDLE_CENTER = 5
    MIDDLE_RIGHT = 6
    TOP_LEFT = 7
    L = 7 # Alias
    TOP_CENTER = 8
    C = 8 # Alias
    TOP_RIGHT = 9
    R = 9 # Alias

@validators.validated_instantiation
class AssPosition(typing.NamedTuple):
    rotation: int
    alignment: AssAlignment
    x: int | float
    y: int | float

    def __str__(self):
        result = "{"
        if self.alignment != AssAlignment.DEFAULT:
            result += r"\an%d" % self.alignment.value
        if self.rotation:
            result += r"\frz%d" % self.rotation
        # Yes, %s, so it will stringify either an int or float and display properly
        result += r"\pos(%s,%s)}" % (self.x, self.y)
        return result

class AssOverflow(enum.Enum):
    NO_WRAP = 2
    EVEN_SPLIT = 0
    TOP_SPLIT = 1
    BOTTOM_SPLIT = 3

    def __str__(self):
        return self.name

@validators.validated_instantiation(replace="__init__")
@dataclasses.dataclass
class AssOptions:
    #position: bool
    #wipe: bool
    border: bool = True # Add CDG-style borders to margins
    #display: int
    #remove: int
    float_font: bool = True # Allow floating point in output font sizes (well-supported)
    float_pos: bool = False # Allow floating point in \pos and margins (supported by recent libass)
    target_x: int = 300 # Output resolution
    target_y: int = 216
    fade_in: int = 300 # Fade duration for line display/remove
    fade_out: int = 200
    transparency: bool = True
    offset: int | bool = True # False = disable offset (same as 0), True = pull from KBS config, int is offset in ms
    overflow: AssOverflow = AssOverflow.EVEN_SPLIT
    allow_kt: bool = False # Use \kt if there are overlapping wipes on the same line (not supported by all ass implementations)
    experimental_spacing: bool = False
    #overflow_spacing: float # TODO? spacing value in styles that will apply for overflow (default 0). Multiplied by font height or based on default style?

    @validators.validated_types
    @staticmethod
    def __assert_valid(key: str, value):
        if key in AssOptions._fields:
            if not isinstance(value, (t := AssOptions._fields[key].type)):
                if callable(t):
                    value = t(value)
                # Also try the first type in a union
                elif hasattr(t, '__args__') and callable(s := t.__args__[0]):
                    value = s(value)
            elif not isinstance(value, t):
                raise TypeError(f"Expected {opt} to be of type {t}. Found {type(options[opt])}.")
        else:
            raise TypeError(f"Unexpected field '{key}'. Possible fields are {self._fields.keys()}.")

        return value

    @validators.validated_structures(assert_function=__assert_valid)
    def update(self, **options):
        for opt in options:
            setattr(self, opt, options[opt])

# Not sure why dataclasses doesn't define something like this keyed by field name
AssOptions._fields = types.MappingProxyType(dict((f.name,f) for f in dataclasses.fields(AssOptions)))

class AssConverter:
    
    @validators.validated_types
    def __init__(self, kbpFile: kbp.KBPFile, options: AssOptions = None, **kwargs):
        self.kbpFile = kbpFile
        # Allow for applying specific overrides to defaults in kwargs, or
        # providing a template for defaults then overriding items there
        self.options = options or AssOptions()
        self.options.update(**kwargs)

    # Delegate to options if not present in the object itself
    def __getattr__(self,attr):
        return getattr(self.options, attr)

    # Move coordinates based on scaling the canvas size
    # If the target aspect ratio is wider than 300:216, x coordinates are
    # scaled more than y and vice versa
    # For example:
    #  - target_x, target_y = 384, 216
    #    - this behaves like -W 384 in the javascript kbp2ass
    #    - Text would not be scaled by scale_font, but this adjusts x positions to expand to fit the space
    #    - y positions would stay the same
    #  - target_x, target_y = 600, 600
    #    - Text would be scaled 2x by scale_font (min(600/300, 600/216))
    #    - x coordinates would be adjusted for the scaling factor
    #    - y coordinates would be further adjusted to fit the space
    #
    # NOTE: Original plan was to provide a letterboxed version as well, but
    # unfortunately .ass cannot reliably do that, as the margins are only used
    # for soft word wrapping. To letterbox, scale to a resolution with aspect
    # ratio the same as CDG, then letterbox when rendering the video
    @validators.validated_types
    @staticmethod
    def rescale_coords(x: int, y: int, target_x: int, target_y: int, allow_float: bool = False, border: bool = True) -> tuple:
        cdg_res = (300, 216) if border else (288, 192)
        res = (x * target_x / cdg_res[0], y * target_y / cdg_res[1])
        if allow_float:
            # Still return ints if they happen to be just as accurate
            return tuple(i if (i := int(r)) == r else r for r in res)
        else:
            return tuple(round(coord) for coord in res)

    # Scale scalars like font, border, shadow
    # These are scaled by a factor of the minimum of the scaling factor of x
    # and y to avoid going off screen
    # If font is true, additional scaling factor of 1.4 is applied to account for line height vs cap height
    @validators.validated_types
    @staticmethod
    def rescale_scalar(size: float, target_x: int, target_y: int, allow_float: bool = True, border: bool = True, font: bool = False) -> int | float:
        cdg_res = (300, 216) if border else (288, 192)
        scale = min(target_x / cdg_res[0], target_y / cdg_res[1])
        res = size * scale * (1.4 if font else 1) # Find way to calculate line_height(fontsize) instead of using this constant
        return res if (allow_float and int(res) != res) else round(res)

    @validators.validated_types
    def get_pos(self, line: kbp.KBPLine, num: int) -> AssPosition:
        cdg_res_x = 300 if self.border else 288
        margins = self.kbpFile.margins
        result = {}
        y = margins["top"] + \
            line.down + \
            num * (self.kbpFile.margins["spacing"] + (kbpfont.spacing(self.kbpFile.styles[1]) if self.experimental_spacing else 19)) + \
            (12 if self.border else 0)

        result["alignment"] = AssAlignment.DEFAULT if line.align == self.style_alignments[line.style] else AssAlignment[line.align]

        if line.align == 'L':
            x = margins["left"] + line.right + (6 if self.border else 0)
        elif line.align == 'C':
            x = cdg_res_x / 2 + line.right
        else: #line.align == 'R' or the file is broken
            x = cdg_res_x - margins["right"] + line.right - (6 if self.border else 0)

        result["x"], result["y"] = AssConverter.rescale_coords(x, y, self.target_x, self.target_y, border=self.border, allow_float=self.float_pos)

        return AssPosition(**result, rotation=line.rotation)

    @validators.validated_types
    def get_line_margins(self, line: kbp.KBPLine, pos: AssPosition | types.NoneType = None, num: int = 1) -> tuple:
        if pos is None:
            pos = self.get_pos(line, num)

        page_margins = self.kbpFile.margins

        left = page_margins["left"] - (0 if line.align == 'L' or (line.align == 'C' and line.right > 0) else line.right)
        right = page_margins["right"] + (0 if line.align == 'R' or (line.align == 'C' and line.right < 0) else line.right)
        left, _ = AssConverter.rescale_coords(left, pos.y, self.target_x, self.target_y, border=self.border, allow_float=self.float_pos)
        right, _ = AssConverter.rescale_coords(right, pos.y, self.target_x, self.target_y, border=self.border, allow_float=self.float_pos)

        return (left, right)
    
    # Determine the most-used line alignment for each style to minimize \anX tags in result
    # (since alignment is not part of the KBP style, but is part of the ASS style)
    def _calc_style_alignments(self):
        # dict of alpha-keyed style to dict of alignment to frequency
        # E.g.
        # { 'A' : {'C': 5, 'L': 2}}
        # would indicate style A was centered 5 times and left-aligned twice
        freqs = collections.defaultdict(lambda: collections.defaultdict(lambda: 0))
        for page in self.kbpFile.pages:
            for line in page.lines:
                freqs[line.style][line.align] += 1
        self.style_alignments = {}
        for style in freqs:
            self.style_alignments[style] = max(freqs[style], key = freqs[style].get)

    @validators.validated_types
    def fade(self) -> str:
        return r"{\fad(%d,%d)}" % (self.options.fade_in, self.options.fade_out)

    # Convert a line of syllables into the text of a dialogue event including wipe tags
    @validators.validated_types
    def kbp2asstext(self, line: kbp.KBPLine, pos: AssPosition):
        result = str(pos) + self.fade()
        if self.kbpFile.styles[line.style].fixed:
            return result + line.text()
        cur = line.start
        for (n, s) in enumerate(line.syllables):
            delay = s.start - cur
            dur = s.end - s.start

            if delay > 0:
                # Gap between current position and start of next syllable
                result += r"{\k%d}" % delay
            elif delay < 0:
                # Playing catchup
                if self.allow_kt:
                    # Reset time so wipes can overlap (\kt takes a time in centiseconds from line start)
                    result += r"{\kt%d}" % (s.start - line.start)
                else:
                    # Shorten syllable to compensate for missing time (keep in mind delay is negative)
                    dur += delay

            # By default a syllable ends 1 centisecond before the next, so
            # special casing so we don't need a bunch of \k1 and the slight
            # errors don't catch up with us on a long line
            if len(line.syllables) > n+1 and line.syllables[n+1].start - s.end == 1:
                dur += 1

            # Using == False explicitly because it's technically a tri-state with None meaning undefined
            # Though that scenario shouldn't come up since we are allowing KBPFile to resolve wipedetail
            wipe = r"\k" if s.isprogressive() == False else r"\kf"

            result += r"{%s%d}%s" % (wipe, dur, s.syllable)
            cur = s.start + dur
        return result

    @validators.validated_types
    @staticmethod
    def ass_style_name(index: int, kbpName: str):
        return f"Style{abs(index):02}_{kbpName}"

    @validators.validated_types(coerce_types=False)
    @staticmethod
    def kbp2asscolor(kbpcolor: int | str, palette: kbp.KBPPalette | types.NoneType = None, transparency: bool = False):
        alpha = "&H00"
        if isinstance(kbpcolor, int):
            if transparency and kbpcolor == 0:
                alpha = "&HFF"
            # This will intentionally raise an exception if colors are unresolved and palette is not provided
            kbpcolor = palette[kbpcolor]
        return alpha + "".join(x+x for x in reversed(list(kbpcolor)))

    def ass_document(self):
        result = ass.Document()
        result.info.update(
            Title="",
            ScriptType="v4.00+",
            WrapStyle=self.overflow.value,
            ScaledBorderAndShadow="yes",
            Collisions="Normal",
            PlayResX=self.options.target_x,
            PlayResY=self.options.target_y,
            ) 

        if self.options.offset is False:
            self.options.offset = 0
        elif self.options.offset is True:
            self.options.offset = kbs.offset * 10
        # else already resolved to an int

        styles = self.kbpFile.styles
        self._calc_style_alignments()
        for page in self.kbpFile.pages:
            for num, line in enumerate(page.lines):
                if line.isempty():
                    continue
                pos = self.get_pos(line, num)
                line_margins = self.get_line_margins(line, pos)
                line_style = styles[line.style]
                result.events.append(ass.Dialogue(
                    start=datetime.timedelta(milliseconds = line.start * 10 + self.options.offset),
                    end=datetime.timedelta(milliseconds = line.end * 10 + self.options.offset),
                    style=AssConverter.ass_style_name(line_style.style_no, line_style.name), # Undefined styles get default style number
                    effect="karaoke",
                    text=self.kbp2asstext(line, pos),
                    margin_l=line_margins[0],
                    margin_r=line_margins[1],
                    ))
        for idx in styles:
            style = styles[idx]
            result.styles.append(ass.Style(
                name=AssConverter.ass_style_name(idx, style.name),
                fontname=style.fontname,
                fontsize=AssConverter.rescale_scalar(style.fontsize, self.target_x, self.target_y, font = True, border=self.border),
                secondary_color=AssConverter.kbp2asscolor(style.textcolor, palette=self.kbpFile.colors, transparency=self.options.transparency),
                primary_color=AssConverter.kbp2asscolor(style.textwipecolor, palette=self.kbpFile.colors, transparency=self.options.transparency),
                outline_color=AssConverter.kbp2asscolor(style.outlinecolor, palette=self.kbpFile.colors, transparency=self.options.transparency),
                # NOTE: no outline wipe in .ass
                back_color=AssConverter.kbp2asscolor(style.outlinewipecolor, palette=self.kbpFile.colors, transparency=self.options.transparency),
                bold = 'B' in style.fontstyle,
                italic = 'I' in style.fontstyle,
                underline = 'U' in style.fontstyle,
                strike_out = 'S' in style.fontstyle,
                outline = AssConverter.rescale_scalar(sum(style.outlines), self.target_x, self.target_y, border=self.border)/4, # NOTE: only one outline, but it's a float, so maybe averaging will be helpful
                shadow = AssConverter.rescale_scalar(sum(style.shadows), self.target_x, self.target_y, border=self.border)/2,
                margin_l = 0, # TODO: Decide if these should be set (and overridden on lines only when different)
                margin_r = 0, # TODO cont: Potentially could also be set by style like alignment based on most-used positions
                margin_v = 0,
                encoding = style.charset,
                alignment=AssAlignment[self.style_alignments.get(kbp.KBPStyleCollection.key2alpha(idx), 'C')].value,
                ))
            
        return result

@validators.validated_instantiation(replace="__init__")
@dataclasses.dataclass
class DoblonTxtOptions:
    title: str = ''
    artist: str = ''
    audio_file: str = ''
    comments: str = 'Created with kbputils\nConverted from Doblon .txt file'
    max_lines_per_page: int = 6
    min_gap_for_new_page: int = 1000
    display_before_wipe: int = 1000
    remove_after_wipe: int = 500
    template_file: str = ''

    @validators.validated_types
    @staticmethod
    def __assert_valid(key: str, value):
        if key in DoblonTxtOptions._fields:
            if not isinstance(value, (t := DoblonTxtOptions._fields[key].type)):
                if callable(t):
                    value = t(value)
                # Also try the first type in a union
                elif hasattr(t, '__args__') and callable(s := t.__args__[0]):
                    value = s(value)
            elif not isinstance(value, t):
                raise TypeError(f"Expected {opt} to be of type {t}. Found {type(options[opt])}.")
        else:
            raise TypeError(f"Unexpected field '{key}'. Possible fields are {self._fields.keys()}.")

        return value

    @validators.validated_structures(assert_function=__assert_valid)
    def update(self, **options):
        for opt in options:
            setattr(self, opt, options[opt])
DoblonTxtOptions._fields = types.MappingProxyType(dict((f.name,f) for f in dataclasses.fields(DoblonTxtOptions)))

class DoblonTxtConverter:
    @validators.validated_types
    def __init__(self, doblonTxtFile: doblontxt.DoblonTxt, options: DoblonTxtOptions | types.NoneType = None, **kwargs):
        self.doblontxt = doblonTxtFile
        self.options = options or DoblonTxtOptions()
        self.options.update(**kwargs)
        self.template = kbp.KBPFile(self.options.template_file, template=True) if self.options.template_file else kbp.KBPFile()

    @staticmethod
    def syl2kbp(syl: str) -> str:
        if syl.endswith("-"):
            syl = syl[:-1]
        return re.sub("/", "{-}", syl)

    def kbpFile(self):
        if hasattr(self, 'kbpfile'):
            return self.kbpfile
        self.kbpfile = self.template
        delattr(self, 'template')
        if not hasattr(self.kbpfile, 'trackinfo'):
            self.kbpfile.trackinfo = {'Status': '1', 'Title': self.options.title, 'Artist': self.options.artist, 'Audio': self.options.audio_file, 'BuildFile': '', 'Intro': '', 'Outro': '', 'Comments': self.options.comments}

        # Add empty line to ensure last page is processed
        self.doblontxt.lines.append([])

        kbplines = []
        for line in self.doblontxt.lines:
            if not line or (kbplines and line[0][1] - self.options.display_before_wipe - kbplines[-1].end*10 > self.options.min_gap_for_new_page):
                if kbplines:
                    num_pages = math.ceil(len(kbplines) / self.options.max_lines_per_page)
                    per_page = len(kbplines) // num_pages
                    additional = len(kbplines) % num_pages
                    cur = 0
                    for group in range(num_pages):
                        nxt = cur + per_page + int(additional > 0)
                        additional -= 1
                        page = kbp.KBPPage("", "", kbplines[cur:nxt])
                        self.kbpfile.pages.append(page)
                        cur = nxt
                    kbplines = []
                if not line:
                    continue
            line_header = kbp.KBPLineHeader(
                align="C",
                style="A",
                start=round((line[0][0] - self.options.display_before_wipe)/10),
                end=round((line[-1][1] + self.options.remove_after_wipe)/10),
                right=0,
                down=0,
                rotation=0
            ) 
            kbpline = kbp.KBPLine(line_header, [kbp.KBPSyllable(self.syl2kbp(syl), round(start/10), round(end/10), 0) for start, end, syl in line])
            kbplines.append(kbpline)
        return self.kbpfile

# TODO: Combine code for LRC and DoblonTXT converters

@validators.validated_instantiation(replace="__init__")
@dataclasses.dataclass
class LRCOptions:
    title: str = ''
    artist: str = ''
    audio_file: str = ''
    comments: str = 'Created with kbputils\nConverted from Enhanced LRC file'
    max_lines_per_page: int = 6
    min_gap_for_new_page: int = 1000
    display_before_wipe: int = 1000
    remove_after_wipe: int = 500
    template_file: str = ''

    @validators.validated_types
    @staticmethod
    def __assert_valid(key: str, value):
        if key in LRCOptions._fields:
            if not isinstance(value, (t := LRCOptions._fields[key].type)):
                if callable(t):
                    value = t(value)
                # Also try the first type in a union
                elif hasattr(t, '__args__') and callable(s := t.__args__[0]):
                    value = s(value)
            elif not isinstance(value, t):
                raise TypeError(f"Expected {opt} to be of type {t}. Found {type(options[opt])}.")
        else:
            raise TypeError(f"Unexpected field '{key}'. Possible fields are {self._fields.keys()}.")

        return value

    @validators.validated_structures(assert_function=__assert_valid)
    def update(self, **options):
        for opt in options:
            setattr(self, opt, options[opt])
LRCOptions._fields = types.MappingProxyType(dict((f.name,f) for f in dataclasses.fields(LRCOptions)))

class LRCConverter:
    @validators.validated_types
    def __init__(self, lrcFile: lrc.LRC, options: LRCOptions | types.NoneType = None, **kwargs):
        self.lrc = lrcFile
        implicitopts = {'ti': 'title', 'ar': 'artist', '#': 'comments'}
        self.options = options or LRCOptions(**{implicitopts[x]: lrcFile.tags[x] for x in lrcFile.tags if x in implicitopts})
        self.options.update(**kwargs)
        self.template = kbp.KBPFile(self.options.template_file, template=True) if self.options.template_file else kbp.KBPFile()

    @staticmethod
    def syl2kbp(syl: str) -> str:
        if syl.endswith("-"):
            syl = syl[:-1]
        return re.sub("/", "{-}", syl)

    def kbpFile(self):
        if hasattr(self, 'kbpfile'):
            return self.kbpfile
        self.kbpfile = self.template
        delattr(self, 'template')
        if not hasattr(self.kbpfile, 'trackinfo'):
            self.kbpfile.trackinfo = {'Status': '1', 'Title': self.options.title, 'Artist': self.options.artist, 'Audio': self.options.audio_file, 'BuildFile': '', 'Intro': '', 'Outro': '', 'Comments': self.options.comments}

        # Add empty line to ensure last page is processed
        self.lrc.lines.append([])

        kbplines = []
        for line in self.lrc.lines:
            if not line or (kbplines and line[0][1] - self.options.display_before_wipe - kbplines[-1].end*10 > self.options.min_gap_for_new_page):
                if kbplines:
                    num_pages = math.ceil(len(kbplines) / self.options.max_lines_per_page)
                    per_page = len(kbplines) // num_pages
                    additional = len(kbplines) % num_pages
                    cur = 0
                    for group in range(num_pages):
                        nxt = cur + per_page + int(additional > 0)
                        additional -= 1
                        page = kbp.KBPPage("", "", kbplines[cur:nxt])
                        self.kbpfile.pages.append(page)
                        cur = nxt
                    kbplines = []
                if not line:
                    continue
            line_header = kbp.KBPLineHeader(
                align="C",
                style="A",
                start=round((line[0][0] - self.options.display_before_wipe)/10),
                end=round((line[-1][1] + self.options.remove_after_wipe)/10),
                right=0,
                down=0,
                rotation=0
            ) 
            kbpline = kbp.KBPLine(line_header, [kbp.KBPSyllable(self.syl2kbp(syl), round(start/10), round(end/10), 0) for start, end, syl in line])
            kbplines.append(kbpline)
        return self.kbpfile



import functools
import inspect

# Decorator for validating the number of arguments passed to a function with a *args parameter
# The assumption is that if the function has a . in its qualified name, it will be used as an
# instance method. To change that behavior, set static to True. By default, unlimited keyword
# arguments are permitted, but if keyword_allowed is set to False, none are
def num_arg(maximum, *, minimum=0, keyword_allowed=True, static=False):
    def num_arg_decorator(func):

        # This is kind of hacky, but there's no way to know if this is going to
        # be a method until an instance is instantiated
        add = 1 if '.' in func.__qualname__ and not static else 0

        if minimum == maximum:
            positional_message = f"exactly {maximum}"
        elif minimum == 0:
            positional_message = f"at most {maximum}"
        else:
            positional_message = f"between {minimum} and {maximum}"
        keyword_message = "" if keyword_allowed else " and no keyword arguments" 

        @functools.wraps(func)
        def num_arg_wrapper(*args, **kwargs):
            if not (minimum + add <= len(args) <= maximum + add) or (not keyword_allowed and kwargs):
                raise TypeError(f"{func.__qualname__} expected {positional_message} positional arguments{keyword_message}, found {len(args) - add} positional and {len(kwargs)} keyword.")
            return func(*args, **kwargs)

        return num_arg_wrapper

    return num_arg_decorator

# Special case of num_arg for maximum=1
def one_arg(func=None, /, **kwargs):
    if func:
        return num_arg(1, **kwargs)(func)
    else:
        return num_arg(1, **kwargs)

# Decorator for checking type signature on a function including return value if annotated
# If coerce_types is set to True, the wrapper will attempt to cast the value to
# the expected type if it doesn't match and send the result to the decorated
# function
def validated_types(func=None, /, *, coerce_types=True):
    def validated_types_generator(func):
        signature = inspect.signature(func)

        @functools.wraps(func)
        def validate_wrapper(*args, **kwargs):
            # Need something mutable if items need to be modified
            if coerce_types:
                args = list(args)
            pos_idx = 0
            for param in signature.parameters:
                comp = None
                setter = None
                if len(args) > pos_idx:
                    comp = args[pos_idx]
                    if coerce_types:
                        setter = (lambda idx: lambda val: args.__setitem__(idx, val))(pos_idx)
                    pos_idx += 1
                elif param in kwargs:
                    comp = kwargs[param]
                    if coerce_types:
                        setter = lambda val: kwargs.__setitem__(param, val)
                else:
                    # parameter not passed, nothing to validate
                    # function itself should validate number of parameters and
                    # whether they were option or mandatory
                    continue 
                if (t := signature.parameters[param].annotation) is not inspect._empty:
                    if coerce_types and not isinstance(comp, t) and callable(t):
                        # Give coercion a try...
                        setter(t(comp))
                    elif not isinstance(comp, t):
                        raise TypeError(f"{func.__qualname__} expected {param} to be of type {t.__name__}, found {type(comp).__name__}.")
            result = func(*args, **kwargs)
            if (t := signature.return_annotation) is not inspect._empty and not isinstance(result, t):
                raise TypeError(f"{func.__qualname__} was expected to return type {t.__name__}, found {type(result).__name__}.")
            return result
        return validate_wrapper
    if func:
        return validated_types_generator(func)
    else:
        return validated_types_generator

# Function decorator for checking an assertion across key/value data sent as a
# single object parameter or provided in kwargs. Ignores the first argument,
# assuming it to be self (or the class) unless static is set to True
# The assert_function passed should take key and value arguments and validate both
# If the assert_function returns a value, it replaces the value passed to the
# decorated function (so types can be coerced instead of just strictly checked)
def validated_structures(assert_function, static=False):
    def validate_structures_decorator(func):
        @functools.wraps(func)
        @one_arg(static=static)
        def validate_structures_wrapper(*args, **kwargs):
            args_tmp = args if static else args[1:]
            arg = args_tmp[0] if args_tmp else []
            to_check = [kwargs]
            if(hasattr(arg,"keys")):
                to_check.append(arg)
            else:
                new_arg = []
                for key,val in arg:
                    res = assert_function(key, val)
                    if res is not None:
                        new_arg.append((key,res))
                    else:
                        new_arg.append((key,val))
                    args = (args[0], new_arg, *args[2:])
            for x in to_check:
                for key in x.keys():
                    res = assert_function(key, x[key])
                    if res is not None:
                        x[key] = res
            return func(*args, **kwargs)
        return validate_structures_wrapper
    return validate_structures_decorator

# Class decorator to validate type annotations during creation of an instance,
# Note that any methods/attributes with type requirements need to be validated
# separately
def validated_instantiation(cls=None, /, *, replace="__new__"):
    def validated_instantiation_generator(cls):
        orig_func = getattr(cls, replace)
        def validator_func(c, *args, **kwargs):
            pos_idx = 0
            for param in (a := inspect.get_annotations(cls)):
                comp = None
                if(len(args) > pos_idx):
                    comp = args[pos_idx]
                    pos_idx += 1
                elif param in kwargs:
                    comp = kwargs[param]
                else:
                    continue
                if not isinstance(comp, (t := a[param])):
                    raise TypeError(f"{cls.__qualname__} expected {param} to be of type {t.__name__}, found {type(comp).__name__}.")
            return orig_func(c, *args, **kwargs)
        setattr(cls, replace, validator_func)
        return cls
    if cls:
        return validated_instantiation_generator(cls)
    else:
        return validated_instantiation_generator
